import pytest
import time
import asyncio
import jsonlines
from costly.costlog import Costlog
from pathlib import Path


def test_costlog_new_item():
    costlog = Costlog(mode="memory", totals_keys={"time"})
    with costlog.new_item() as (item, timer):
        time.sleep(0.35)
        item.update({"cost": "Hello", "time": timer()})
    assert costlog.items == [{"cost": "Hello", "time": pytest.approx(0.35, abs=0.02)}]
    assert costlog.totals["time"] == pytest.approx(0.35, abs=0.02)


def test_costlog_new_item_multiple():
    costlog = Costlog(mode="memory", totals_keys={"time"})
    with costlog.new_item() as (item, timer):
        time.sleep(0.08)
        item.update({"cost": "Hello", "time": timer()})
    with costlog.new_item() as (item, timer):
        time.sleep(0.12)
        item.update({"cost": "Hello", "time": timer()})
    assert costlog.items == [
        {"cost": "Hello", "time": pytest.approx(0.08, abs=0.02)},
        {"cost": "Hello", "time": pytest.approx(0.12, abs=0.02)},
    ]
    assert costlog.totals["time"] == pytest.approx(0.20, abs=0.02)


def test_costlog_jsonl():
    costlog = Costlog(
        mode="jsonl",
        totals_keys={"time"},
        path=Path("tests", "test_costlog.jsonl"),
        overwrite=True,
    )
    with costlog.new_item() as (item, timer):
        time.sleep(0.08)
        item.update({"cost": "Hello", "time": timer()})
    with costlog.new_item() as (item, timer):
        time.sleep(0.12)
        item.update({"cost": "Hello", "time": timer()})
    assert costlog.totals["time"] == pytest.approx(0.20, abs=0.01)
    with open(costlog.path, "r") as f:
        items = [item for item in jsonlines.Reader(f)]
    assert items == [
        {"cost": "Hello", "time": pytest.approx(0.08, abs=0.01)},
        {"cost": "Hello", "time": pytest.approx(0.12, abs=0.01)},
    ]


@pytest.mark.asyncio
async def test_costlog_new_item_async():
    costlog = Costlog(mode="memory", totals_keys={"time"})
    async with costlog.new_item_async() as (item, timer):
        await asyncio.sleep(0.35)
        item.update({"cost": "Hello", "time": timer()})
    assert costlog.items == [{"cost": "Hello", "time": pytest.approx(0.35, abs=0.02)}]
    assert costlog.totals["time"] == pytest.approx(0.35, abs=0.02)


@pytest.mark.asyncio
async def test_costlog_new_item_async_multiple():
    costlog = Costlog(mode="memory", totals_keys={"time"})
    async with costlog.new_item_async() as (item, timer):
        await asyncio.sleep(0.08)
        item.update({"cost": "Hello", "time": timer()})
    async with costlog.new_item_async() as (item, timer):
        await asyncio.sleep(0.12)
        item.update({"cost": "Hello", "time": timer()})
    assert costlog.items == [
        {"cost": "Hello", "time": pytest.approx(0.08, abs=0.02)},
        {"cost": "Hello", "time": pytest.approx(0.12, abs=0.02)},
    ]
    assert costlog.totals["time"] == pytest.approx(0.20, abs=0.02)


@pytest.mark.asyncio
async def test_costlog_jsonl_async():
    costlog = Costlog(
        mode="jsonl",
        totals_keys={"time"},
        path=Path("tests", "test_costlog_async.jsonl"),
        overwrite=True,
    )
    async with costlog.new_item_async() as (item, timer):
        await asyncio.sleep(0.08)
        item.update({"cost": "Hello", "time": timer()})
    async with costlog.new_item_async() as (item, timer):
        await asyncio.sleep(0.12)
        item.update({"cost": "Hello", "time": timer()})
    assert costlog.totals["time"] == pytest.approx(0.20, abs=0.02)
    with open(costlog.path, "r") as f:
        items = [item for item in jsonlines.Reader(f)]
    assert items == [
        {"cost": "Hello", "time": pytest.approx(0.08, abs=0.02)},
        {"cost": "Hello", "time": pytest.approx(0.12, abs=0.02)},
    ]


@pytest.mark.asyncio
async def test_costlog_async_parallel():
    costlog = Costlog(mode="memory", totals_keys={"time"})

    async def task(sleep_time):
        async with costlog.new_item_async() as (item, timer):
            await asyncio.sleep(sleep_time)
            item.update({"time": timer()})

    await asyncio.gather(task(0.1), task(0.2), task(0.3))

    assert len(costlog.items) == 3
    assert costlog.items[0]["time"] == pytest.approx(0.1, abs=0.02)
    assert costlog.items[1]["time"] == pytest.approx(0.2, abs=0.03)
    assert costlog.items[2]["time"] == pytest.approx(0.3, abs=0.04)
    assert costlog.totals["time"] == pytest.approx(0.6, abs=0.05)


@pytest.mark.asyncio
async def test_costlog_async_mixed():
    costlog = Costlog(mode="memory", totals_keys={"time"})

    with costlog.new_item() as (item, timer):
        time.sleep(0.1)
        item.update({"time": timer()})

    async with costlog.new_item_async() as (item, timer):
        await asyncio.sleep(0.2)
        item.update({"time": timer()})

    assert len(costlog.items) == 2
    assert costlog.items[0] == {"time": pytest.approx(0.1, abs=0.02)}
    assert costlog.items[1] == {"time": pytest.approx(0.2, abs=0.03)}
    assert costlog.totals["time"] == pytest.approx(0.3, abs=0.03)



import instructor
from typing import Any
from pydantic import BaseModel
from costly import costly, CostlyResponse
from costly.estimators.llm_api_estimation import LLM_API_Estimation
from costly.simulators.llm_simulator_faker import LLM_Simulator_Faker
from openai import OpenAI, AsyncOpenAI
from instructor import Instructor
import asyncio

class PERSONINFO(BaseModel):
    name: str
    age: int


class FOOMODEL(BaseModel):
    name: str
    age: int
    bmi: float
    metadata: dict[str, Any] | None = None


class BARMODEL(BaseModel):
    foo: FOOMODEL
    fookids: list[FOOMODEL]


CLIENT = instructor.from_openai(OpenAI())
CLIENT_ASYNC = instructor.from_openai(AsyncOpenAI())

@costly()
def chatgpt(messages: list[dict[str, str]], model: str) -> str:
    from openai import OpenAI

    client = OpenAI()
    response = client.chat.completions.create(model=model, messages=messages)
    return CostlyResponse(
        output=response.choices[0].message.content,
        cost_info={
            "input_tokens": response.usage.prompt_tokens,
            "output_tokens": response.usage.completion_tokens,
        },
    )


@costly(
    messages=(lambda kwargs: kwargs["history"]),
    model=(lambda kwargs: kwargs["model_name"]),
)
def chatgpt2(history: list[dict[str, str]], model_name: str) -> str:
    from openai import OpenAI

    client = OpenAI()
    response = client.chat.completions.create(model=model_name, messages=history)
    return CostlyResponse(
        output=response.choices[0].message.content,
        cost_info={
            "input_tokens": response.usage.prompt_tokens,
            "output_tokens": response.usage.completion_tokens,
        },
    )


@costly(messages="history", model="model_name")
def chatgpt3(history: list[dict[str, str]], model_name: str) -> str:
    from openai import OpenAI

    client = OpenAI()
    response = client.chat.completions.create(model=model_name, messages=history)
    return CostlyResponse(
        output=response.choices[0].message.content,
        cost_info={
            "input_tokens": response.usage.prompt_tokens,
            "output_tokens": response.usage.completion_tokens,
        },
    )


@costly(
    input_tokens=lambda kwargs: LLM_API_Estimation.prompt_to_input_tokens(**kwargs),
)
def chatgpt_prompt(
    prompt: str, model: str, system_prompt: str = "You are a helpful assistant."
) -> str:
    from openai import OpenAI

    client = OpenAI()
    response = client.chat.completions.create(
        model=model,
        messages=[
            {"content": prompt, "role": "user"},
            {"content": system_prompt, "role": "system"},
        ],
    )
    output_string = response.choices[0].message.content
    return output_string


@costly(
    input_tokens=lambda kwargs: LLM_API_Estimation.get_input_tokens_instructor(
        **kwargs
    ),
)
def chatgpt_instructor(
    messages: list[dict[str, str]],
    model: str,
    client: Instructor,
    response_model: BaseModel,
) -> str:
    response = client.chat.completions.create_with_completion(
        model=model,
        messages=messages,
        response_model=response_model,
    )
    output_string, cost_info = response
    return CostlyResponse(
        output=output_string,
        cost_info={
            "input_tokens": cost_info.usage.prompt_tokens,
            "output_tokens": cost_info.usage.completion_tokens,
        },
    )


@costly()
async def chatgpt_async(messages: list[dict[str, str]], model: str) -> str:
    from openai import AsyncOpenAI

    client = AsyncOpenAI()
    response = await client.chat.completions.create(model=model, messages=messages)
    return CostlyResponse(
        output=response.choices[0].message.content,
        cost_info={
            "input_tokens": response.usage.prompt_tokens,
            "output_tokens": response.usage.completion_tokens,
        },
    )


@costly(
    messages=(lambda kwargs: kwargs["history"]),
    model=(lambda kwargs: kwargs["model_name"]),
)
async def chatgpt2_async(history: list[dict[str, str]], model_name: str) -> str:
    from openai import AsyncOpenAI

    client = AsyncOpenAI()
    response = await client.chat.completions.create(model=model_name, messages=history)
    return CostlyResponse(
        output=response.choices[0].message.content,
        cost_info={
            "input_tokens": response.usage.prompt_tokens,
            "output_tokens": response.usage.completion_tokens,
        },
    )


@costly(
    input_tokens=lambda kwargs: LLM_API_Estimation.prompt_to_input_tokens(**kwargs),
)
async def chatgpt_prompt_async(
    prompt: str, model: str, system_prompt: str = "You are a helpful assistant."
) -> str:
    from openai import AsyncOpenAI

    client = AsyncOpenAI()
    response = await client.chat.completions.create(
        model=model,
        messages=[
            {"content": prompt, "role": "user"},
            {"content": system_prompt, "role": "system"},
        ],
    )
    output_string = response.choices[0].message.content
    return output_string


@costly(
    input_tokens=lambda kwargs: LLM_API_Estimation.get_input_tokens_instructor(
        **kwargs
    ),
)
async def chatgpt_instructor_async(
    messages: list[dict[str, str]],
    model: str,
    client: Instructor,
    response_model: BaseModel,
) -> str:
    response = await client.chat.completions.create_with_completion(
        model=model,
        messages=messages,
        response_model=response_model,
    )
    output_string, cost_info = response
    return CostlyResponse(
        output=output_string,
        cost_info={
            "input_tokens": cost_info.usage.prompt_tokens,
            "output_tokens": cost_info.usage.completion_tokens,
        },
    )


@costly(
    simulator=LLM_Simulator_Faker.simulate_llm_probs,
)
def chatgpt_probs(messages: list[dict[str, str]], model: str, return_probs_for: list[str]) -> dict[str, float]:
    from openai import OpenAI

    client = OpenAI()
    response = client.chat.completions.create(
        model=model,
        messages=messages,
        logprobs=True,
        top_logprobs=len(return_probs_for)
    )
    
    logprobs = response.choices[0].logprobs.content[0].top_logprobs
    probs = {option: 0 for option in return_probs_for}
    
    for logprob in logprobs:
        if logprob.token in return_probs_for:
            probs[logprob.token] = 2 ** logprob.logprob  # Convert log probability to probability
    
    # Normalize probabilities
    total = sum(probs.values())
    normalized_probs = {k: v / total for k, v in probs.items()}
    
    return CostlyResponse(
        output=normalized_probs,
        cost_info={
            "input_tokens": response.usage.prompt_tokens,
            "output_tokens": response.usage.completion_tokens,
        },
    )



from types import NoneType, GenericAlias, UnionType
from typing import Any, Optional, Union, get_origin, get_args, Sequence, Mapping
import json
from pydantic import BaseModel


def isinstance_better(v, t: type) -> bool:

    origin = get_origin(t)
    args = get_args(t)

    if origin in (Union, UnionType):
        return any(isinstance_better(v, arg) for arg in get_args(t))

    if origin is None:
        return isinstance(v, t)
    if not isinstance(v, origin):
        return False
    if isinstance(v, Sequence) and not isinstance(v, (str, bytes)):
        return all(isinstance_better(item, args[0]) for item in v)
    elif isinstance(v, Mapping):
        return all(
            isinstance_better(k, args[0]) and isinstance_better(v, args[1])
            for k, v in v.items()
        )
    return None


def json_serializable(value):
    try:
        json.dumps(value)
        return True
    except (TypeError, ValueError):
        return False


def make_json_serializable(value):
    if isinstance(value, dict):
        return {k: make_json_serializable(v) for k, v in value.items()}
    elif isinstance(value, list):
        return [make_json_serializable(v) for v in value]
    elif isinstance(value, tuple):
        return tuple(make_json_serializable(v) for v in value)
    elif isinstance(value, set):
        return set(make_json_serializable(v) for v in value)
    elif isinstance(value, BaseModel):
        return make_json_serializable(value.model_dump(mode="json"))
    elif not json_serializable(value):
        return str(value)
    return value



import pytest
from types import NoneType
from typing import Optional, Union, Any
from pydantic import BaseModel
from costly.simulators.llm_simulator_faker import LLM_Simulator_Faker
from costly.utils import isinstance_better
from costly.costlog import Costlog
from tests.example_functions import FOOMODEL, BARMODEL

TYPES = {
    int,
    str,
    float,
    list,
    dict,
    tuple,
    set,
    list[str],
    list[int],
    list[float],
    dict[str, int],
    dict[str, float],
    dict[str, Optional[list]] | str | None,
    FOOMODEL,
    list[FOOMODEL],
    BARMODEL,
    NoneType,
    Union[int, str],
    float | str,
}


@pytest.mark.parametrize("t", TYPES)
def test_fake_type(t: type):
    x = LLM_Simulator_Faker.fake(t)
    assert isinstance_better(x, t), f"Expected {t}, got {type(x)}"


def test_simulate_llm_probs():
    return_probs_for = ["option1", "option2", "option3"]
    costlog = Costlog()
    
    probs = LLM_Simulator_Faker.simulate_llm_probs(
        return_probs_for=return_probs_for,
        input_string="Test input",
        model="gpt-3.5-turbo",
        cost_log=costlog
    )
    
    assert isinstance(probs, dict)
    assert set(probs.keys()) == set(return_probs_for)
    assert sum(probs.values()) == pytest.approx(1.0)
    
    for value in probs.values():
        assert 0 <= value <= 1
    
    assert len(costlog.items) == 1
    cost_item = costlog.items[0]
    assert "input_tokens" in cost_item
    assert "output_tokens_min" in cost_item
    assert "output_tokens_max" in cost_item
    assert cost_item["output_tokens_min"] == 0
    assert cost_item["output_tokens_max"] == max(len(key) for key in return_probs_for)



"""
Tests to make sure the estimates for simulation actually match the actual cost.
"""

import pytest
from openai import OpenAI
import instructor
from pydantic import BaseModel
from costly.costlog import Costlog
from costly.estimators.llm_api_estimation import LLM_API_Estimation
from tests.example_functions import (
    chatgpt,
    chatgpt_prompt,
    chatgpt_instructor,
    chatgpt_probs,
    CLIENT,
    PERSONINFO,
    FOOMODEL,
    BARMODEL,
)

PROMPTS = [
    "Explain the Riemann Hypothesis in full mathematical detail.",
    "Write a short story.",
    "Write the bubblesort algorithm in Python.",
]
MESSAGESS = [[{"content": prompt, "role": "user"}] for prompt in PROMPTS]
MODELS = ["gpt-4o", "gpt-4o-mini", "gpt-3.5-turbo", "claude-3-5-sonnet-20240620", "claude-3-5-sonnet-20241022"]
MODELS_OPENAI = [model for model in MODELS if model.startswith("gpt")]
PARAMSS = [
    {"messages": messages, "model": model}
    for model in MODELS_OPENAI
    for messages in MESSAGESS
]
PARAMSS_PROMPT = [
    {"prompt": prompt, "model": model} for model in MODELS_OPENAI for prompt in PROMPTS
]
PROMPTS_INSTRUCTOR = [
    "Give me an example of a PERSONINFO object.",
    "Give me an example of a FOOMODEL object.",
    "Give me an example of a BARMODEL object.",
]
MESSAGESS_INSTRUCTOR = [
    [{"content": prompt, "role": "user"}] for prompt in PROMPTS_INSTRUCTOR
]
PARAMSS_INSTRUCTOR_ = {
    "PERSONINFO": {"messages": MESSAGESS_INSTRUCTOR[0], "response_model": PERSONINFO},
    "FOOMODEL": {"messages": MESSAGESS_INSTRUCTOR[1], "response_model": FOOMODEL},
    "BARMODEL": {"messages": MESSAGESS_INSTRUCTOR[2], "response_model": BARMODEL},
}
PARAMSS_INSTRUCTOR = {
    k + "_" + model: {**v, "model": model, "client": instructor.from_openai(OpenAI())}
    for k, v in PARAMSS_INSTRUCTOR_.items()
    for model in MODELS_OPENAI
}

@pytest.mark.parametrize("model", MODELS)
def check_obvious_things(model):
    assert model in LLM_API_Estimation.PRICES
    assert LLM_API_Estimation.get_model(model) == model
    all_matches = LLM_API_Estimation.get_model(model, return_all_matches=True)
    assert isinstance(all_matches, list)
    assert all_matches[0] == model
    assert all_matches[-1] == "__default__"
    prices = LLM_API_Estimation.get_prices(model)
    assert isinstance(prices, dict)
    assert "input_cost_per_token" in prices
    assert "output_cost_per_token" in prices
    assert "time" in prices
    if model in LLM_API_Estimation.TIMES:
        assert prices["time"] == LLM_API_Estimation.TIMES[model]
    if model.startswith("claude"): # temporary thing
        assert prices["time"] == LLM_API_Estimation.TIMES["__default__"]

def check_cost_estimates(real_cost, sim_cost):
    print("real_cost", real_cost)
    print("sim_cost", sim_cost)
    assertions = [
        {
            "check": 0.8 * real_cost["input_tokens"]
            <= sim_cost["input_tokens"]
            <= 1.2 * real_cost["input_tokens"],
            "message": f"Input tokens estimate {sim_cost['input_tokens']} not within 20pc of truth {real_cost['input_tokens']}",
        },
        {
            "check": sim_cost["output_tokens_min"] <= real_cost["output_tokens_min"],
            "message": f"Output tokens estimate minimum {sim_cost['output_tokens_min']} exceeds truth {real_cost['output_tokens_min']}",
        },
        {
            "check": real_cost["output_tokens_max"] <= sim_cost["output_tokens_max"],
            "message": f"Output tokens estimate maximum {sim_cost['output_tokens_max']} is less than truth {real_cost['output_tokens_max']}",
        },
        {
            "check": sim_cost["time_min"] <= real_cost["time_min"],
            "message": f"Time estimate minimum {sim_cost['time_min']} exceeds truth {real_cost['time_min']}",
        },
        {
            "check": real_cost["time_max"] <= sim_cost["time_max"],
            "message": f"Time estimate maximum {sim_cost['time_max']} is less than truth {real_cost['time_max']}",
        },
        {
            "check": sim_cost["cost_min"] <= real_cost["cost_min"],
            "message": f"Cost estimate minimum {sim_cost['cost_min']} exceeds truth {real_cost['cost_min']}",
        },
        {
            "check": real_cost["cost_max"] <= sim_cost["cost_max"],
            "message": f"Cost estimate maximum {sim_cost['cost_max']} is less than truth {real_cost['cost_max']}",
        },
    ]
    failures = [assertion for assertion in assertions if not assertion["check"]]
    assert not failures, [assertion["message"] for assertion in failures]


@pytest.mark.slow
@pytest.mark.parametrize("messages", MESSAGESS)
@pytest.mark.parametrize("model", MODELS_OPENAI)
def test_estimate_contains_exact(messages, model):
    costlog = Costlog()
    real = chatgpt(
        messages=messages,
        model=model,
        simulate=False,
        cost_log=costlog,
    )
    sim = chatgpt(
        messages=messages,
        model=model,
        simulate=True,
        cost_log=costlog,
    )
    real_cost = costlog.items[0]
    sim_cost = costlog.items[1]
    check_cost_estimates(real_cost, sim_cost)


@pytest.mark.slow
@pytest.mark.parametrize("prompt", PROMPTS)
@pytest.mark.parametrize("model", MODELS_OPENAI)
def test_estimate_contains_exact_prompt(prompt, model):
    costlog = Costlog()
    real = chatgpt_prompt(
        prompt=prompt,
        model=model,
        simulate=False,
        cost_log=costlog,
    )
    sim = chatgpt_prompt(
        prompt=prompt,
        model=model,
        simulate=True,
        cost_log=costlog,
    )
    real_cost = costlog.items[0]
    sim_cost = costlog.items[1]
    check_cost_estimates(real_cost, sim_cost)


@pytest.mark.slow
@pytest.mark.parametrize("params", PARAMSS_INSTRUCTOR.keys())
def test_estimate_contains_exact_instructor(params):
    costlog = Costlog()
    real = chatgpt_instructor(
        **PARAMSS_INSTRUCTOR[params],
        simulate=False,
        cost_log=costlog,
    )
    sim = chatgpt_instructor(
        **PARAMSS_INSTRUCTOR[params],
        simulate=True,
        cost_log=costlog,
    )
    real_cost = costlog.items[0]
    sim_cost = costlog.items[1]
    check_cost_estimates(real_cost, sim_cost)


@pytest.mark.slow
def test_estimate_contains_exact_probs():
    return_probs_for = [str(n) for n in range(10)]
    costlog = Costlog()

    messages = [
        {
            "content": (
                "Take a random guess as to what the 1,000,001st digit of pi is."
                'Answer exactly "0", "1", ... or "9", with nothing else in your response.'
            ),
            "role": "user",
        }
    ]
    model = "gpt-4o"

    real = chatgpt_probs(
        messages=messages,
        model=model,
        return_probs_for=return_probs_for,
        simulate=False,
        cost_log=costlog,
    )

    sim = chatgpt_probs(
        messages=messages,
        model=model,
        return_probs_for=return_probs_for,
        simulate=True,
        cost_log=costlog,
    )

    real_cost = costlog.items[0]
    sim_cost = costlog.items[1]
    check_cost_estimates(real_cost, sim_cost)



import pytest
import asyncio
import instructor
from pydantic import BaseModel
from costly import costly, Costlog
from costly.estimators.llm_api_estimation import LLM_API_Estimation
from openai import OpenAI, AsyncOpenAI
from instructor import Instructor
from tests.example_functions import (
    chatgpt,
    chatgpt2,
    chatgpt3,
    chatgpt_prompt,
    chatgpt_instructor,
    chatgpt_async,
    chatgpt2_async,
    chatgpt_prompt_async,
    chatgpt_instructor_async,
    CLIENT,
    CLIENT_ASYNC,
    PERSONINFO,
)


def test_chatgpt():
    costlog = Costlog()
    x = chatgpt(
        messages=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model="gpt-4o-mini",
        cost_log=costlog,
        simulate=False,
        description=["chatgpt call"],
    )
    y = chatgpt(
        messages=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model="gpt-4o-mini",
        cost_log=costlog,
        simulate=True,
        description=["chatgpt call"],
    )
    assert isinstance(x, str)
    assert isinstance(y, str)
    assert len(costlog.items) == 2
    assert costlog.totals["calls"] == 2


def test_chatgpt2():
    costlog = Costlog()
    x = chatgpt2(
        history=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model_name="gpt-4o-mini",
        cost_log=costlog,
        simulate=False,
        description=["chatgpt call"],
    )
    y = chatgpt2(
        history=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model_name="gpt-4o-mini",
        cost_log=costlog,
        simulate=True,
        description=["chatgpt call"],
    )
    assert isinstance(x, str)
    assert isinstance(y, str)
    assert len(costlog.items) == 2
    assert costlog.totals["calls"] == 2


def test_chatgpt3():
    costlog = Costlog()
    x = chatgpt3(
        history=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model_name="gpt-4o-mini",
        cost_log=costlog,
        simulate=False,
        description=["chatgpt call"],
    )
    y = chatgpt3(
        history=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model_name="gpt-4o-mini",
        cost_log=costlog,
        simulate=True,
        description=["chatgpt call"],
    )
    assert isinstance(x, str)
    assert isinstance(y, str)
    assert len(costlog.items) == 2
    assert costlog.totals["calls"] == 2


def test_chatgpt_prompt():
    costlog = Costlog()
    x = chatgpt_prompt(
        prompt="Write the Lorem ipsum text",
        model="gpt-4o-mini",
        cost_log=costlog,
        simulate=False,
        description=["chatgpt call"],
    )
    y = chatgpt_prompt(
        prompt="Write the Lorem ipsum text",
        model="gpt-4o-mini",
        cost_log=costlog,
        simulate=True,
        description=["chatgpt call"],
    )
    assert isinstance(x, str)
    assert isinstance(y, str)
    assert len(costlog.items) == 2
    assert costlog.totals["calls"] == 2


def test_chatgpt_instructor():
    costlog = Costlog()
    x = chatgpt_instructor(
        messages=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model="gpt-4o-mini",
        client=CLIENT,
        response_model=PERSONINFO,
        cost_log=costlog,
        simulate=False,
        description=["chatgpt call"],
    )
    y = chatgpt_instructor(
        messages=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model="gpt-4o-mini",
        client=CLIENT,
        response_model=PERSONINFO,
        cost_log=costlog,
        simulate=True,
        description=["chatgpt call"],
    )
    assert isinstance(x, PERSONINFO)
    assert isinstance(y, PERSONINFO)
    assert len(costlog.items) == 2
    assert costlog.totals["calls"] == 2


@pytest.mark.asyncio
async def test_chatgpt_async():
    costlog = Costlog()
    x = await chatgpt_async(
        messages=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model="gpt-4o-mini",
        cost_log=costlog,
        simulate=False,
        description=["chatgpt call"],
    )
    y = await chatgpt_async(
        messages=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model="gpt-4o-mini",
        cost_log=costlog,
        simulate=True,
        description=["chatgpt call"],
    )
    assert isinstance(x, str)
    assert isinstance(y, str)
    assert len(costlog.items) == 2
    assert costlog.totals["calls"] == 2


@pytest.mark.asyncio
async def test_chatgpt2_async():
    costlog = Costlog()
    x = await chatgpt2_async(
        history=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model_name="gpt-4o-mini",
        cost_log=costlog,
        simulate=False,
        description=["chatgpt call"],
    )
    y = await chatgpt2_async(
        history=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model_name="gpt-4o-mini",
        cost_log=costlog,
        simulate=True,
        description=["chatgpt call"],
    )
    assert isinstance(x, str)
    assert isinstance(y, str)
    assert len(costlog.items) == 2
    assert costlog.totals["calls"] == 2


@pytest.mark.asyncio
async def test_chatgpt_prompt_async():
    costlog = Costlog()
    x = await chatgpt_prompt_async(
        prompt="Write the Lorem ipsum text",
        model="gpt-4o-mini",
        cost_log=costlog,
        simulate=False,
        description=["chatgpt call"],
    )
    y = await chatgpt_prompt_async(
        prompt="Write the Lorem ipsum text",
        model="gpt-4o-mini",
        cost_log=costlog,
        simulate=True,
        description=["chatgpt call"],
    )
    assert isinstance(x, str)
    assert isinstance(y, str)
    assert len(costlog.items) == 2
    assert costlog.totals["calls"] == 2


@pytest.mark.asyncio
async def test_chatgpt_instructor_async():
    costlog = Costlog()
    x = await chatgpt_instructor_async(
        messages=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model="gpt-4o-mini",
        client=CLIENT_ASYNC,
        response_model=PERSONINFO,
        cost_log=costlog,
        simulate=False,
        description=["chatgpt call"],
    )
    y = await chatgpt_instructor_async(
        messages=[{"role": "user", "content": "Write the Lorem ipsum text"}],
        model="gpt-4o-mini",
        client=CLIENT_ASYNC,
        response_model=PERSONINFO,
        cost_log=costlog,
        simulate=True,
        description=["chatgpt call"],
    )
    assert isinstance(x, PERSONINFO)
    assert isinstance(y, PERSONINFO)
    assert len(costlog.items) == 2
    assert costlog.totals["calls"] == 2


@pytest.mark.asyncio
async def test_chatgpt_async_parallel():
    costlog = Costlog()
    messages = [
        [{"role": "user", "content": "Write a short story"}],
        [{"role": "user", "content": "Explain quantum computing"}],
        [{"role": "user", "content": "Describe the water cycle"}],
    ]

    results = await asyncio.gather(
        *[
            chatgpt_async(
                messages=msg, model="gpt-4o-mini", cost_log=costlog, simulate=False
            )
            for msg in messages
        ]
    )

    assert len(results) == 3
    assert all(isinstance(result, str) for result in results)
    assert len(costlog.items) == 3
    assert costlog.totals["calls"] == 3


@pytest.mark.asyncio
async def test_chatgpt2_async_parallel():
    costlog = Costlog()
    histories = [
        [{"role": "user", "content": "Write a haiku"}],
        [{"role": "user", "content": "Explain photosynthesis"}],
        [{"role": "user", "content": "Describe the solar system"}],
    ]

    results = await asyncio.gather(
        *[
            chatgpt2_async(
                history=hist, model_name="gpt-4o-mini", cost_log=costlog, simulate=False
            )
            for hist in histories
        ]
    )

    assert len(results) == 3
    assert all(isinstance(result, str) for result in results)
    assert len(costlog.items) == 3
    assert costlog.totals["calls"] == 3


@pytest.mark.asyncio
async def test_chatgpt_prompt_async_parallel():
    costlog = Costlog()
    prompts = [
        "Write a limerick",
        "Explain the theory of relativity",
        "Describe the process of photosynthesis",
    ]

    results = await asyncio.gather(
        *[
            chatgpt_prompt_async(
                prompt=prompt, model="gpt-4o-mini", cost_log=costlog, simulate=False
            )
            for prompt in prompts
        ]
    )

    assert len(results) == 3
    assert all(isinstance(result, str) for result in results)
    assert len(costlog.items) == 3
    assert costlog.totals["calls"] == 3


@pytest.mark.asyncio
async def test_chatgpt_instructor_async_parallel():
    costlog = Costlog()
    messages_list = [
        [{"role": "user", "content": "Generate a person named Alice"}],
        [{"role": "user", "content": "Generate a person named Bob"}],
        [{"role": "user", "content": "Generate a person named Charlie"}],
    ]

    results = await asyncio.gather(
        *[
            chatgpt_instructor_async(
                messages=msgs,
                model="gpt-4o-mini",
                client=CLIENT_ASYNC,
                response_model=PERSONINFO,
                cost_log=costlog,
                simulate=False,
            )
            for msgs in messages_list
        ]
    )

    assert len(results) == 3
    assert all(isinstance(result, PERSONINFO) for result in results)
    assert len(costlog.items) == 3
    assert costlog.totals["calls"] == 3


@pytest.mark.asyncio
async def test_mixed_async_parallel():
    costlog = Costlog()

    results = await asyncio.gather(
        chatgpt_async(
            messages=[{"role": "user", "content": "Write a short story"}],
            model="gpt-4o-mini",
            cost_log=costlog,
            simulate=False,
        ),
        chatgpt2_async(
            history=[{"role": "user", "content": "Explain quantum computing"}],
            model_name="gpt-4o-mini",
            cost_log=costlog,
            simulate=False,
        ),
        chatgpt_prompt_async(
            prompt="Describe the water cycle",
            model="gpt-4o-mini",
            cost_log=costlog,
            simulate=False,
        ),
        chatgpt_instructor_async(
            messages=[{"role": "user", "content": "Generate a person named David"}],
            model="gpt-4o-mini",
            client=CLIENT_ASYNC,
            response_model=PERSONINFO,
            cost_log=costlog,
            simulate=False,
        ),
    )

    assert len(results) == 4
    assert isinstance(results[0], str)
    assert isinstance(results[1], str)
    assert isinstance(results[2], str)
    assert isinstance(results[3], PERSONINFO)
    assert len(costlog.items) == 4
    assert costlog.totals["calls"] == 4



from costly import Costlog, CostlyResponse, costly
from dotenv import load_dotenv

load_dotenv()  # remove when done

cl = Costlog()
nosim = False
sim = True


@costly()
def chatgpt(
    messages: list[dict[str, str]],
    model: str,
    cost_log: Costlog = cl,
    simulate: bool = nosim,
) -> str:
    from openai import OpenAI

    client = OpenAI()
    response = client.chat.completions.create(model=model, messages=messages)
    return CostlyResponse(
        output=response.choices[0].message.content,
        cost_info={
            "input_tokens": response.usage.prompt_tokens,
            "output_tokens": response.usage.completion_tokens,
        },
    )


@costly()
def chatgpt_simulated(
    messages: list[dict[str, str]],
    model: str,
    cost_log: Costlog = cl,
    simulate: bool = sim,
) -> str:
    from openai import OpenAI

    client = OpenAI()
    response = client.chat.completions.create(model=model, messages=messages)
    return CostlyResponse(
        output=response.choices[0].message.content,
        cost_info={
            "input_tokens": response.usage.prompt_tokens,
            "output_tokens": response.usage.completion_tokens,
        },
    )


def test_global():
    x = chatgpt(
        messages=[{"content": "Write the Lorem ipsum text", "role": "user"}],
        model="gpt-4o-mini",
    )
    y = chatgpt_simulated(
        messages=[{"content": "Write the Lorem ipsum text", "role": "user"}],
        model="gpt-4o-mini",
    )
    assert len(cl.items) == 2
    assert cl.items[0]["simulated"] == False
    assert cl.items[1]["simulated"] == True



import os
import datetime
import json
import jsonlines
import time
import asyncio
from dataclasses import dataclass
from typing import Literal
from contextlib import contextmanager, asynccontextmanager
from pathlib import Path
from costly.utils import make_json_serializable


class Costlog(dict):

    def __init__(
        self,
        path: Path | str = None,
        mode: Literal["jsonl", "memory"] = "memory",
        totals_keys: set[str] = None,
        discard_extras: bool = True,
        overwrite: bool = None,
    ):
        """
        Arguments:
            path: Path to the costlog file, if in jsonl mode. If None, the costlog will be stored in
                the current working directory in a folder called ".costly".
            mode: "jsonl" or "memory". "jsonl" will store the costlog in a jsonl file,
                "memory" will store the costlog in memory which may not be suitable for large logs.
            totals_keys: The keys to store in the totals.
            discard_extras: If True, extra keys in the costlog items will be discarded.
            overwrite: If the costlog file already exists, overwrite it. If None, the user will be asked. Only relevant if mode is "jsonl".
        """
        match mode:
            case "jsonl":
                if path is None:
                    self.path = Path(
                        os.getcwd(),
                        ".costly",
                        f"costlog_{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.jsonl",
                    )
                else:
                    self.path = Path(path)
                if self.path.exists():
                    if overwrite is None:
                        overwrite = input(
                            f"Costlog file {self.path} already exists. Overwrite? (y/n) "
                        )
                    if overwrite not in ["y", True]:
                        raise FileExistsError(
                            f"Costlog file {self.path} already exists."
                        )
                    else:
                        self.path.unlink()
                else:
                    self.path.parent.mkdir(parents=True, exist_ok=True)
                    with open(self.path, "w") as f:
                        jsonlines.Writer(f)
                self.totals_path = self.path.with_suffix(".totals.json")
                self.totals_by_model_path = self.path.with_suffix(
                    ".totals_by_model.json"
                )
            case "memory":
                self.items = []
            case _:
                raise ValueError(f"Invalid mode: {mode}")
        self.mode = mode
        if totals_keys is None:
            totals_keys = {
                "cost_min",
                "cost_max",
                "time_min",
                "time_max",
                "calls",
                "input_tokens",
                "output_tokens_min",
                "output_tokens_max",
            }
        self.totals_keys = totals_keys
        self.acceptable_keys = totals_keys | {"model", "simulated", "cost", "time"}
        self.discard_extras = discard_extras
        self.totals = {key: 0.0 for key in totals_keys}
        self.totals_by_model = {}
        super().__init__(self.totals)

    def append(self, **kwargs):
        match self.mode:
            case "jsonl":
                with jsonlines.open(self.path, "a") as f:
                    f.write(make_json_serializable(kwargs))
            case "memory":
                self.items.append(kwargs)
        for key in self.totals:
            self.totals[key] += kwargs.get(key, 0.0)
            if kwargs.get("model", "__UNKNOWN__") not in self.totals_by_model:
                self.totals_by_model[kwargs.get("model", "__UNKNOWN__")] = {
                    key: 0.0 for key in self.totals_keys
                }
            self.totals_by_model[kwargs.get("model", "__UNKNOWN__")][key] += kwargs.get(
                key, 0.0
            )
        if self.mode == "jsonl":
            with open(self.totals_path, "w") as f:
                json.dump(self.totals, f, indent=4)
            with open(self.totals_by_model_path, "w") as f:
                json.dump(self.totals_by_model, f, indent=4)

    @contextmanager
    def new_item(self):
        item = {}
        t1 = time.perf_counter()
        yield item, lambda: time.perf_counter() - t1
        if self.discard_extras:
            item = {key: item[key] for key in item if key in self.acceptable_keys}
        self.append(**item)

    @asynccontextmanager
    async def new_item_async(self):
        item = {}
        t1 = time.perf_counter()
        yield item, lambda: time.perf_counter() - t1
        if self.discard_extras:
            item = {key: item[key] for key in item if key in self.acceptable_keys}
        self.append(**item)

    def totals_from_items(self):
        match self.mode:
            case "jsonl":
                with jsonlines.open(self.path) as f:
                    items = list(f)
            case "memory":
                items = self.items
            case _:
                raise ValueError(f"Invalid mode: {self.mode}")
        return {key: sum(item[key] for item in items) for key in self.totals_keys}
